from photonai_conformal.mapie_wrapper import ConformalClassifier, ConformalRegressor
from sklearn.gaussian_process import GaussianProcessRegressor, GaussianProcessClassifier
from sklearn.datasets import load_diabetes, load_iris
import numpy as np
from mapie.classification import MapieClassifier
from mapie.regression import MapieRegressor
from unittest import TestCase


class ConformalTests(TestCase):

    def test_regressor_basic(self):
        X, y = load_diabetes(return_X_y=True)
        reg = ConformalRegressor(estimator=GaussianProcessRegressor()).fit(X, y)
        baseline = MapieRegressor(estimator=GaussianProcessRegressor()).fit(X, y)

        reg_preds = reg.predict(X)
        baseline_preds = baseline.predict(X)
        self.assertTrue(np.array_equal(reg_preds, baseline_preds))

    def test_regressor_advanced_nonequal(self):
        X, y = load_diabetes(return_X_y=True)
        reg = ConformalRegressor(estimator=GaussianProcessRegressor(), alpha=[.5]).fit(X, y)
        baseline = MapieRegressor(estimator=GaussianProcessRegressor()).fit(X, y)

        reg_preds = reg.predict(X)
        baseline_preds = baseline.predict(X, alpha=[.5])
        self.assertTrue(np.array_equal(reg_preds['y_pred'], baseline_preds[0]))
        self.assertTrue(np.array_equal(reg_preds['0.5_lower'], baseline_preds[1][:, 0, 0]))
        self.assertTrue(np.array_equal(reg_preds['0.5_upper'], baseline_preds[1][:, 1, 0]))

    def test_classifier_basic(self):
        X, y = load_iris(return_X_y=True)
        novel = ConformalClassifier(estimator=GaussianProcessClassifier()).fit(X, y)
        baseline = MapieClassifier(estimator=GaussianProcessClassifier()).fit(X, y)

        novel_preds = novel.predict(X)
        baseline_preds = baseline.predict(X)
        self.assertTrue(np.array_equal(novel_preds, baseline_preds))

    def test_classifier_advanced_nonequal(self):
        X, y = load_iris(return_X_y=True)
        novel = ConformalClassifier(estimator=GaussianProcessClassifier(),
                                    alpha=[.2, .3, .5, .9],
                                    random_state=43,
                                    method='naive').fit(X, y)

        novel_preds, y_pred, y_pis = novel.predict(X, return_internals_testing=True)
        self.assertTrue(np.array_equal(novel_preds['y_pred'], y_pred))
        self.assertTrue(np.array_equal(novel_preds['0.2_0_ps'], y_pis[:, 0, 0]))
        self.assertTrue(np.array_equal(novel_preds['0.2_1_ps'], y_pis[:, 1, 0]))



import os
from datetime import datetime
from photonai.base import PhotonRegistry
from photonai.photonlogger import logger

from .version import __version__


def do_register(current_path, registered_file): # pragma: no cover
    reg = PhotonRegistry()
    reg.add_module(os.path.join(current_path, "photonai_conformal.json"))
    with open(os.path.join(registered_file), "w") as f:
        f.write(str(__version__))


def register(): # pragma: no cover
    current_path = os.path.dirname(os.path.abspath(__file__))
    registered_file = os.path.join(current_path, "registered")
    logger.info("Checking photonai_conformal Module Registration")
    if not os.path.isfile(registered_file):  # pragma: no cover
        logger.info("Registering photonai_conformal Module")
        do_register(current_path=current_path, registered_file=registered_file)
    else:
        with open(os.path.join(registered_file), "r") as f:
            if f.read() == __version__:
                logger.info("Current version already registered")
            else:
                logger.info("Updating photonai_conformal Module")
                do_register(current_path=current_path, registered_file=registered_file)


register()



from abc import ABC
import numpy as np
import itertools
from sklearn.base import BaseEstimator, RegressorMixin, ClassifierMixin

from mapie.classification import MapieClassifier
from mapie.regression import MapieRegressor


class MapieWrapper(ABC, BaseEstimator):

    def __init__(self):
        self.clf = NotImplementedError
        self.alpha = None

    def fit(self, X, y):
        self.clf.fit(X, y)
        return self

    def predict(self, X, **kwargs):
        raise NotImplementedError()


class ConformalClassifier(MapieWrapper, ClassifierMixin):
    def __init__(self,
                 estimator=None,
                 method="score",
                 cv=None,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 alpha=None):
        """
        Wrapper for conformal classification.For parameter details see
        [mapie docs](https://mapie.readthedocs.io/en/latest/generated/mapie.classification.MapieClassifier.html#mapie-classification-mapieclassifier)
        Parameters
        ----------
        estimator
        method
        cv
        n_jobs
        random_state
        verbose
        alpha
        """
        super(ConformalClassifier, self).__init__()
        self.estimator = estimator
        self.method = method
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.alpha = alpha
        self.clf = MapieClassifier(estimator=estimator,
                                   method=method,
                                   cv=cv,
                                   n_jobs=n_jobs,
                                   random_state=random_state,
                                   verbose=verbose)

    def predict(self, X, **kwargs):
        if not self.alpha and 'alpha' not in kwargs:
            return self.clf.predict(X)
        else:
            alpha_vals = self.alpha.copy() if self.alpha else kwargs['alpha']
            preds_orig, y_pis = self.clf.predict(X, alpha=alpha_vals)
            preds = np.stack([preds_orig[:]]*y_pis.shape[1], axis=-1)[..., None]  # dims
            output = np.append(y_pis, preds, axis=-1)  # preds are at [..., -1]
            alpha_vals.append('y_pred')
            final_preds = np.array([tuple(output[row, c_ps, alpha] for alpha, c_ps in
                                          itertools.product(range(len(alpha_vals)), range(output.shape[1]))) for row in
                                    range(output.shape[0])],
                                   dtype=[(str(key) + ul, np.float64) for key, ul in
                                          itertools.product(alpha_vals, [f'_{i}_ps' for i in range(output.shape[1])])])
            # remove double y_pred entry
            pred_names = list(final_preds.dtype.names[:-2])
            final_preds = final_preds[pred_names]
            pred_names = pred_names[:-1]
            pred_names.append('y_pred')
            final_preds.dtype.names = tuple(pred_names)
            if 'return_internals_testing' in kwargs:
                return final_preds, preds_orig, y_pis
            return final_preds


class ConformalRegressor(MapieWrapper, RegressorMixin):
    def __init__(self,
                 estimator=None,
                 method="plus",
                 cv=None,
                 n_jobs=None,
                 agg_function="mean",
                 verbose=0,
                 conformity_score=None,
                 alpha=None):
        """
        Wrapper for conformal regression. For parameter details see
        [mapie docs](https://mapie.readthedocs.io/en/latest/generated/mapie.regression.MapieRegressor.html#mapie.regression.MapieRegressor)
        Parameters
        ----------
        estimator
        method
        cv
        n_jobs
        agg_function
        verbose
        conformity_score
        alpha
        """
        super(ConformalRegressor, self).__init__()
        self.estimator = estimator
        self.method = method
        self.cv = cv
        self.n_jobs = n_jobs
        self.agg_function = agg_function
        self.verbose = verbose
        self.conformity_score = conformity_score
        self.alpha = alpha
        self.clf = MapieRegressor(estimator=estimator,
                                  method=method,
                                  cv=cv,
                                  n_jobs=n_jobs,
                                  agg_function=agg_function,
                                  verbose=verbose,
                                  conformity_score=conformity_score)

    def predict(self, X, **kwargs):
        if not self.alpha and 'alpha' not in kwargs:
            return self.clf.predict(X)
        else:
            alpha_vals = self.alpha.copy() if self.alpha else kwargs['alpha']
            preds, y_pis = self.clf.predict(X, alpha=alpha_vals)
            preds = np.stack((preds[:], preds[:]), axis=-1)[..., None]  # dims
            output = np.append(y_pis, preds, axis=-1)  # preds are at [..., -1]
            alpha_vals.append('y_pred')
            final_preds = np.array([tuple(output[row, upper_lower, alpha] for alpha, upper_lower in
                            itertools.product(range(0, len(alpha_vals)), [0, 1])) for row in range(output.shape[0])],
                     dtype=[(str(key) + ul, np.float64) for key, ul in
                            itertools.product(alpha_vals, ['_lower', '_upper'])])
            # remove double y_pred entry
            pred_names = list(final_preds.dtype.names[:-1])
            final_preds = final_preds[pred_names]
            pred_names = pred_names[:-1]
            pred_names.append('y_pred')
            final_preds.dtype.names = tuple(pred_names)
            return final_preds
